# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2016 and onwards Google, Inc.
#
# See www.openfst.org for extensive documentation on this weighted
# finite-state transducer library.

import os
from typing import Optional, Union, overload, TypeVar, Tuple, Any, Type, Iterable, Iterator, List

## Custom exceptions.
class FstError(Exception): ...
class FstArgError(FstError, ValueError): ...
class FstBadWeightError(FstError, ValueError): ...
class FstDeletedConstructorError(FstError, RuntimeError): ...
class FstIndexError(FstError, IndexError): ...
class FstIOError(FstError, IOError): ...
class FstOpError(FstError, RuntimeError): ...

# Custom Types
_Filename = Union[os.PathLike, str]
_FarFileModeFlag = str
_ArcTypeFlag = str
_FarTypeFlag = str
_Symbol = str
_Label = int
_StateId = int
_FstTypeFlag = str
_SortTypeFlag = str
# TODO(wolfsonkin): Replace with a typing.Literal[] once Python 3.8 hits.
_QueueType = str
# TODO(wolfsonkin): Replace with a typing.Literal[] once Python 3.8 hits.
_ComposeFilterFlag = str

## Weight and helpers.

_WeightType = str
_WeightStr = Union[str, int, float]
WeightLike = Union[Weight, _WeightStr]

class Weight:
  def __repr__(self) -> str: ...
  def __str__(self) -> str: ...
  def __float__(self) -> float: ...
  def __init__(self, weight_type: _WeightType, weight: _WeightStr) -> None: ...
  def copy(self) -> Weight: ...
  @classmethod
  def Zero(cls, weight_type: _WeightType) -> Weight: ...
  @classmethod
  def One(cls, weight_type: _WeightType) -> Weight: ...
  @classmethod
  def NoWeight(cls, weight_type: _WeightType) -> Weight: ...
  def __eq__(w1: Weight, w2: Weight) -> bool: ...
  def __ne__(w1: Weight, w2: Weight) -> bool: ...
  def to_string(self) -> str: ...
  def type(self) -> str: ...
  def member(self) -> bool: ...

# TODO(wolfsonkin): Look into converting these to operator overloading instead.
def plus(lhs: Weight, rhs: Weight) -> Weight: ...
def times(lhs: Weight, rhs: Weight) -> Weight: ...
def divide(lhs: Weight, rhs: Weight) -> Weight: ...
def power(w: Weight, n: int) -> Weight: ...
class SymbolTableView:
  def __init__(self) -> None: ...
  def __iter__(self) -> _SymbolTableIterator: ...
  def __reduce__(self) -> Union[str, Tuple[Any, ...]]: ...
  def available_key(self) -> _Label: ...
  def checksum(self) -> bytes: ...
  # NOTE: This is intentional. `copy()` produces a underscore-less (owned) SymbolTable object.
  def copy(self) -> SymbolTable: ...
  @overload
  def find(self, key: _Symbol) -> _Label: ...
  @overload
  def find(self, key: _Label) -> _Symbol: ...
  def get_nth_key(self, pos: int) -> _Label: ...
  def labeled_checksum(self) -> bytes: ...
  def member(self, key: Union[_Symbol, _Label]) -> bool: ...
  def __contains__(self, key: Union[_Symbol, _Label]) -> bool: ...
  def name(self) -> str: ...
  def num_symbols(self) -> int: ...
  def write(self, source: _Filename) -> None: ...
  def write_text(self, source: _Filename) -> None: ...
  def write_to_string(self) -> bytes: ...

class _EncodeMapperSymbolTableView(SymbolTableView):
  def __repr__(self) -> str: ...

class _FstSymbolTableView(SymbolTableView):
  def __repr__(self) -> str: ...

class _MutableSymbolTable(SymbolTableView):
  def add_symbol(self, symbol: _Symbol, key: _Label = ...) -> _Label: ...
  def add_table(self, syms: SymbolTableView) -> None: ...
  def set_name(self, new_name: str) -> None: ...

class _MutableFstSymbolTableView(_MutableSymbolTable):
  def __repr__(self) -> str: ...

class SymbolTable(_MutableSymbolTable):
  def __repr__(self) -> str: ...
  def __init__(self, name: str = ...) -> None: ...
  @classmethod
  def read(cls, source: _Filename) -> SymbolTable: ...
  @classmethod
  def read_text(cls,
                source: _Filename,
                allow_negative_labels: bool = ...) -> SymbolTable: ...
  @classmethod
  def read_fst(cls, source: _Filename, input_table: bool) -> SymbolTable: ...

# Constructive SymbolTable operations.

def compact_symbol_table(syms: SymbolTableView) -> SymbolTable: ...
def merge_symbol_table(lhs: SymbolTableView,
                       rhs: SymbolTableView) -> SymbolTable: ...

## _SymbolTableIterator.

class _SymbolTableIterator:
  def __repr__(self) -> str: ...
  def __init__(self, syms: SymbolTableView) -> None: ...
  def __iter__(self) -> _SymbolTableIterator: ...
  def __next__(self) -> Tuple[_Label, _Symbol]: ...

## EncodeMapper.

class EncodeMapper:
  def __repr__(self) -> None: ...
  def __init__(self,
               arc_type: _ArcTypeFlag = ...,
               encode_labels: bool = ...,
               encode_weights: bool = ...) -> None: ...
  def __call__(self, arc: Arc) -> Arc: ...
  def __reduce__(self) -> Union[str, Tuple[Any, ...]]: ...
  def arc_type(self) -> _ArcTypeFlag: ...
  def weight_type(self) -> _WeightType: ...
  def flags(self) -> int: ...
  def properties(self, mask: int) -> int: ...
  @classmethod
  def read(cls, source: _Filename) -> EncodeMapper: ...
  @staticmethod
  def read_from_string(state: bytes) -> EncodeMapper: ...
  def write(self, source: _Filename) -> None: ...
  def write_to_string(self) -> bytes: ...
  def input_symbols(self) -> _EncodeMapperSymbolTableView: ...
  def output_symbols(self) -> _EncodeMapperSymbolTableView: ...
  def set_input_symbols(self,
                        syms: Optional[SymbolTableView]) -> EncodeMapper: ...
  def set_output_symbols(self,
                         syms: Optional[SymbolTableView]) -> EncodeMapper: ...

U = TypeVar("U", bound="Fst")
class Fst:
  # TODO(wolfsonkin): Verify that `bytes` is really the output type here.
  def _repr_svg_(self) -> bytes: ...
  def __init__(self) -> None: ...
  def __reduce__(self) -> Union[str, Tuple[Any, ...]]: ...
  def __repr__(self) -> str: ...
  def __str__(self) -> str: ...
  def arc_type(self) -> _ArcTypeFlag: ...
  def arcs(self, state: _StateId) -> ArcIterator: ...
  def copy(self: U) -> U: ...
  def draw(
      self,
      source: _Filename,
      isymbols: Optional[SymbolTableView] = ...,
      osymbols: Optional[SymbolTableView] = ...,
      ssymbols: Optional[SymbolTableView] = ...,
      acceptor: bool = ...,
      title: str = ...,
      width: float = ...,
      height: float = ...,
      portrait: bool = ...,
      vertical: bool = ...,
      ranksep: float = ...,
      nodesep: float = ...,
      fontsize: int = ...,
      precision: int = ...,
      # TODO(wolfsonkin): Replace this with typing.Literal when Python 3.8 hits.
      float_format: str = ...,
      show_weight_one: bool = ...) -> None: ...
  def final(self, state: _StateId) -> Weight: ...
  def fst_type(self) -> _FstTypeFlag: ...
  def input_symbols(self) -> _FstSymbolTableView: ...
  def num_arcs(self, state: _StateId) -> int: ...
  def num_input_epsilons(self, state: _StateId) -> int: ...
  def num_output_epsilons(self, state: _StateId) -> int: ...
  def output_symbols(self) -> _FstSymbolTableView: ...
  def print(self,
            isymbols: Optional[SymbolTableView] = ...,
            osymbols: Optional[SymbolTableView] = ...,
            ssymbols: Optional[SymbolTableView] = ...,
            acceptor: bool = ...,
            show_weight_one: bool = ...,
            missing_sym: _Symbol = ...) -> str: ...
  def properties(self, mask: int, test: bool) -> int: ...
  @classmethod
  def read(cls: Type[U], source: _Filename) -> U: ...
  @classmethod
  def read_from_string(cls: Type[U], state: bytes) -> U: ...
  def start(self) -> _StateId: ...
  def states(self) -> StateIterator: ...
  def text(self,
           isymbols: Optional[SymbolTableView] = ...,
           osymbols: Optional[SymbolTableView] = ...,
           ssymbols: Optional[SymbolTableView] = ...,
           acceptor: bool = ...,
           show_weight_one: bool = ...,
           missing_sym: _Symbol = ...) -> str: ...
  def verify(self) -> bool: ...
  def weight_type(self) -> _WeightType: ...
  def write(self, source: _Filename) -> None: ...
  def write_to_string(self) -> bytes: ...

T = TypeVar("T", bound="MutableFst")
class MutableFst(Fst):
  def add_arc(self: T, state: _StateId, arc: Arc) -> T: ...
  def add_state(self) -> _StateId: ...
  def add_states(self, n: int) -> None: ...
  def arcsort(self: T, sort_type: _SortTypeFlag = ...) -> T: ...
  def closure(self: T, closure_plus: bool = ...) -> T: ...
  def concat(self: T, fst2: Fst) -> T: ...
  def connect(self: T) -> T: ...
  def decode(self: T, mapper: EncodeMapper) -> T: ...
  def delete_arcs(self: T, state: _StateId, n: int = ...) -> T: ...
  def delete_states(self: T, states: Optional[Iterable[_StateId]] = ...) -> T: ...
  def encode(self: T, mapper: EncodeMapper) -> T: ...
  def invert(self: T) -> T: ...
  def minimize(self: T, delta: float = ..., allow_nondet: bool = ...) -> T: ...
  def mutable_arcs(self, state: _StateId) -> MutableArcIterator: ...
  def mutable_input_symbols(self) -> _MutableFstSymbolTableView: ...
  def mutable_output_symbols(self) -> _MutableFstSymbolTableView: ...
  def num_states(self) -> _StateId: ...
  def project(self: T, project_output: bool = ...) -> T: ...
  def prune(self: T,
            delta: float = ...,
            nstate: _StateId = ...,
            weight: Optional[WeightLike] = ...) -> T: ...
  def push(self: T,
           delta: float = ...,
           remove_total_weight: bool = ...,
           to_final: bool = ...) -> T: ...
  def relabel_pairs(
      self: T,
      ipairs: Optional[Iterable[Tuple[_Label, _Label]]] = ...,
      opairs: Optional[Iterable[Tuple[_Label, _Label]]] = ...) -> T: ...
  def relabel_tables(self: T,
                     old_isymbols: Optional[SymbolTableView] = ...,
                     new_isymbols: Optional[SymbolTableView] = ...,
                     unknown_isymbol: _Symbol = ...,
                     attach_new_isymbols: bool = ...,
                     old_osymbols: Optional[SymbolTableView] = ...,
                     new_osymbols: Optional[SymbolTableView] = ...,
                     unknown_osymbol: _Symbol = ...,
                     attach_new_osymbols: bool = ...) -> T: ...
  def reserve_arcs(self: T, state: _StateId, n: int) -> T: ...
  def reserve_states(self: T, n: _StateId) -> T: ...
  def reweight(self: T,
               potentials: Iterable[WeightLike],
               to_final: bool = ...) -> T: ...
  def rmepsilon(self: T,
                queue_type: _QueueType = ...,
                connect: bool = ...,
                weight: Optional[WeightLike] = ...,
                nstate: _StateId = ...,
                delta: float = ...) -> T: ...
  def set_final(self: T,
                state: _StateId,
                weight: Optional[WeightLike] = ...) -> T: ...
  def set_input_symbols(self: T, syms: Optional[SymbolTableView]) -> T: ...
  def set_output_symbols(self: T, syms: Optional[SymbolTableView]) -> T: ...
  def set_properties(self: T, props: int, mask: int) -> T: ...
  def set_start(self: T, state: _StateId) -> T: ...
  def topsort(self: T) -> T: ...
  def union(self: T, *fsts2: Fst) -> T: ...

class VectorFst(MutableFst):
  def __init__(self, arc_type: _ArcTypeFlag = ...) -> None: ...

def _read_Fst(source: _Filename) -> Fst: ...
def _read_Fst_from_string(state: str) -> Fst: ...
NO_LABEL: _Label
NO_STATE_ID: _StateId
NO_SYMBOL: _Symbol

## FST properties.

EXPANDED: int
MUTABLE: int
ERROR: int
ACCEPTOR: int
NOT_ACCEPTOR: int
I_DETERMINISTIC: int
NON_I_DETERMINISTIC: int
O_DETERMINISTIC: int
NON_O_DETERMINISTIC: int
EPSILONS: int
NO_EPSILONS: int
I_EPSILONS: int
NO_I_EPSILONS: int
O_EPSILONS: int
NO_O_EPSILONS: int
I_LABEL_SORTED: int
NOT_I_LABEL_SORTED: int
O_LABEL_SORTED: int
NOT_O_LABEL_SORTED: int
WEIGHTED: int
UNWEIGHTED: int
CYCLIC: int
ACYCLIC: int
INITIAL_CYCLIC: int
INITIAL_ACYCLIC: int
TOP_SORTED: int
NOT_TOP_SORTED: int
ACCESSIBLE: int
NOT_ACCESSIBLE: int
COACCESSIBLE: int
NOT_COACCESSIBLE: int
STRING: int
NOT_STRING: int
WEIGHTED_CYCLES: int
UNWEIGHTED_CYCLES: int
NULL_PROPERTIES: int
COPY_PROPERTIES: int
INTRINSIC_PROPERTIES: int
EXTRINSIC_PROPERTIES: int
SET_START_PROPERTIES: int
SET_FINAL_PROPERTIES: int
ADD_STATE_PROPERTIES: int
ADD_ARC_PROPERTIES: int
SET_ARC_PROPERTIES: int
DELETE_STATE_PROPERTIES: int
DELETE_ARC_PROPERTIES: int
STATE_SORT_PROPERTIES: int
ARC_SORT_PROPERTIES: int
I_LABEL_INVARIANT_PROPERTIES: int
O_LABEL_INVARIANT_PROPERTIES: int
WEIGHT_INVARIANT_PROPERTIES: int
ADD_SUPERFINAL_PROPERTIES: int
RM_SUPERFINAL_PROPERTIES: int
BINARY_PROPERTIES: int
TRINARY_PROPERTIES: int
POS_TRINARY_PROPERTIES: int
NEG_TRINARY_PROPERTIES: int
FST_PROPERTIES: int

## Arc iterator properties.

ARC_I_LABEL_VALUE: int
ARC_O_LABEL_VALUE: int
ARC_WEIGHT_VALUE: int
ARC_NEXT_STATE_VALUE: int
ARC_NO_CACHE: int
ARC_VALUE_FLAGS: int
ARC_FLAGS: int

## EncodeMapper properties.

ENCODE_LABELS: int
ENCODE_WEIGHTS: int
ENCODE_FLAGS: int

## Arc, ArcIterator, and MutableArcIterator.

class Arc:
  def __repr__(self) -> str: ...
  def __init__(self, ilabel: _Label, olabel: _Label,
               weight: WeightLike, nextstate: _StateId) -> None: ...
  def copy(self) -> Arc: ...
  @property
  def ilabel(self) -> _Label: ...
  @ilabel.setter
  def ilabel(self, value: _Label) -> None: ...
  @property
  def olabel(self) -> _Label: ...
  @olabel.setter
  def olabel(self, value: _Label) -> None: ...
  @property
  def weight(self) -> Weight: ...
  @weight.setter
  def weight(self, value: WeightLike) -> None: ...
  @property
  def nextstate(self) -> _Label: ...
  @nextstate.setter
  def nextstate(self, value: _Label) -> None: ...

class ArcIterator:
  def __repr__(self) -> str: ...
  # TODO(wolfsonkin): Consider making this _FstOrStr instead.
  def __init__(self, ifst: Fst, state: _StateId) -> None: ...
  def __iter__(self) -> ArcIterator: ...
  def __next__(self) -> Arc: ...
  def done(self) -> bool: ...
  def flags(self) -> int: ...
  def next(self) -> None: ...
  def position(self) -> int: ...
  def reset(self) -> None: ...
  def seek(self, a: int) -> None: ...
  def set_flags(self, flags: int, mask: int) -> None: ...
  def value(self) -> Arc: ...

class MutableArcIterator:
  def __repr__(self) -> str: ...
  # TODO(wolfsonkin): Consider making this _FstOrStr instead.
  def __init__(self, ifst: MutableFst, state: _StateId) -> None: ...
  def __iter__(self) -> Iterator[Arc]: ...
  def done(self) -> bool: ...
  def flags(self) -> int: ...
  def next(self) -> None: ...
  def position(self) -> int: ...
  def reset(self) -> None: ...
  def seek(self, a: int) -> None: ...
  def set_flags(self, flags: int, mask: int) -> None: ...
  def set_value(self, arc: Arc) -> None: ...
  def value(self) -> Arc: ...

## StateIterator.

class StateIterator:
  def __repr__(self) -> str: ...
  # TODO(wolfsonkin): Consider making this _FstOrStr instead.
  def __init__(self, ifst: Fst) -> None: ...
  def __iter__(self) -> ArcIterator: ...
  def __next__(self) -> Arc: ...
  def done(self) -> bool: ...
  def next(self) -> None: ...
  def reset(self) -> None: ...
  def value(self) -> _StateId: ...

## FST operations.


def arcmap(
    ifst: Fst,
    delta: float = ...,
    # TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
    map_type: str = ...,
    power: float = ...,
    weight: Optional[WeightLike] = ...) -> Fst: ...
def compose(
    ifst1: Fst,
    ifst2: Fst,
    compose_filter: _ComposeFilterFlag = ...,
    connect: bool = ...) -> MutableFst: ...
def convert(ifst: Fst, fst_type: _FstTypeFlag = ...) -> Fst: ...
def determinize(
    ifst: Fst,
    delta: float = ...,
    # TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
    det_type: str = ...,
    nstate: _StateId = ...,
    subsequential_label: _Label = ...,
    weight: Optional[WeightLike] = ...,
    increment_subsequential_label: bool = ...) -> MutableFst: ...
def difference(
    ifst1: Fst,
    ifst2: Fst,
    compose_filter: _ComposeFilterFlag = ...,
    connect: bool = ...) -> MutableFst: ...
def disambiguate(ifst: Fst,
                 delta: float = ...,
                 nstate: _StateId = ...,
                 subsequential_label: _Label = ...,
                 weight: Optional[WeightLike] = ...) -> MutableFst: ...
def epsnormalize(ifst: Fst, eps_norm_output: bool = ...) -> MutableFst: ...
def equal(ifst1: Fst, ifst2: Fst, delta: float = ...) -> bool: ...
def equivalent(ifst1: Fst, ifst2: Fst, delta: float = ...) -> bool: ...
def intersect(ifst1: Fst,
              ifst2: Fst,
              compose_filter: _ComposeFilterFlag = ...,
              connect: bool = ...) -> MutableFst: ...
def isomorphic(ifst1: Fst, ifst2: Fst, delta: float = ...) -> bool: ...
def prune(ifst: Fst,
          delta: float = ...,
          nstate: _StateId = ...,
          weight: Optional[WeightLike] = ...) -> MutableFst: ...
def push(ifst: Fst,
         delta: float = ...,
         push_weights: bool = ...,
         push_labels: bool = ...,
         remove_common_affix: bool = ...,
         remove_total_weight: bool = ...,
         to_final: bool = ...) -> MutableFst: ...
def randequivalent(
    ifst1: Fst,
    ifst2: Fst,
    npath: int = ...,
    delta: float = ...,
    # TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
    select: str = ...,
    max_length: int = ...,
    seed: int = ...) -> bool: ...
def randgen(
    ifst: Fst,
    npath: int = ...,
    # TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
    select: str = ...,
    max_length: int = ...,
    weighted: bool = ...,
    remove_total_weight: bool = ...,
    seed: int = ...) -> MutableFst: ...
def replace(
    pairs: Iterable[Tuple[int, Fst]],
    # TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
    call_arc_labeling: str = ...,
    # TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
    return_arc_labeling: str = ...,
    epsilon_on_replace: bool = ...,
    return_label: _Label = ...) -> MutableFst: ...
def reverse(ifst: Fst, require_superinitial: bool = ...) -> MutableFst: ...
def shortestdistance(
    ifst: Fst,
    delta: float = ...,
    nstate: _StateId = ...,
    queue_type: _QueueType = ...,
    reverse: bool = ...) -> List[Weight]: ...
def shortestpath(
    ifst: Fst,
    delta: float = ...,
    nshortest: int = ...,
    nstate: _StateId = ...,
    queue_type: _QueueType = ...,
    unique: bool = ...,
    weight: Optional[WeightLike] = ...) -> MutableFst: ...
# TODO(wolfsonkin): Use typing.Literal when Python 3.8 hits.
def statemap(ifst: Fst, map_type: str) -> Fst: ...
def synchronize(ifst: Fst) -> MutableFst: ...
## Compiler.

class Compiler:
  # NOTE: Though the actual function in cython is __cinit__, it's exposed to
  # python as __init__.
  def __init__(
      self,
      fst_type: _FstTypeFlag = ...,
      arc_type: _ArcTypeFlag = ...,
      # TODO(wolfsonkin): Look into converting these to SymbolTableView.
      isymbols: Optional[SymbolTable] = ...,
      osymbols: Optional[SymbolTable] = ...,
      ssymbols: Optional[SymbolTable] = ...,
      acceptor: bool = ...,
      keep_isymbols: bool = ...,
      keep_osymbols: bool = ...,
      keep_state_numbering: bool = ...,
      allow_negative_labels: bool = ...) -> None: ...
  def compile(self) -> Fst: ...
  def write(self, expression: str) -> None: ...

## FarReader and FarWriter.

class FarReader:
  def __init__(self) -> None: ...
  def __repr__(self) -> str: ...
  @classmethod
  def open(cls, *sources: _Filename) -> FarReader: ...
  def arc_type(self) -> _ArcTypeFlag: ...
  def done(self) -> bool: ...
  def error(self) -> bool: ...
  def far_type(self) -> _FarTypeFlag: ...
  def find(self, key: str) -> bool: ...
  def get_fst(self) -> Fst: ...
  def get_key(self) -> str: ...
  def next(self) -> None: ...
  def reset(self) -> None: ...
  def __getitem__(self, key: str) -> Fst: ...

class FarWriter:
  def __init__(self) -> None: ...
  def __repr__(self) -> str: ...
  @classmethod
  def create(cls,
             source: _Filename,
             arc_type: _ArcTypeFlag = ...,
             far_type: _FarTypeFlag = ...) -> FarWriter: ...
  def add(self, key: str, ifst: Fst) -> None: ...
  def arc_type(self) -> _ArcTypeFlag: ...
  def error(self) -> bool: ...
  def far_type(self) -> _FarTypeFlag: ...
  def __setitem__(self, key: str, fst: Fst) -> None: ...

